/*
    $Id: main.cpp 180 2011-01-30 15:56:33Z matwey $
    Copyright (C) 2025 Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <cctype>
#include <iostream>
#include <stdexcept>

#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/program_options.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include <yalpa/sun.h>
#include <yalpa/sidtime.h>

#include "atmos.h" // for VERSION
#include "average.h" // for utils::algo
#include "background_factory.h" // for atmos::background_factory
#include "file_input.h" // for atmos::input::file
#include "file_output.h" // for atmos::output::file
#include "indices.h" // for atmos::indices::*
#include "io.h"
#include "local_context.h" // for utils::local_context
#include "params.h" // for atmos::parameters::*
#include "table_storage.h" // for table_storage
#include "typesdef.h" // for datum
#include "ublas_math.h" // for utils::ublas::log

namespace prephot {

constexpr const double mag_ln_const = 1.08573620476;

namespace flux {
	typedef atmos::indices::time_type time_type;
	typedef time_type::time_duration_type time_duration_type;
	typedef atmos::indices::flux_type flux_type;
	typedef double airmass_type;

namespace detail {
	struct sample {
		typedef prephot::flux::time_type time_type;
		typedef prephot::flux::time_duration_type time_duration_type;
		typedef prephot::flux::flux_type flux_type;
		typedef prephot::flux::airmass_type airmass_type;

		time_duration_type time;
		flux_type    flux;
		airmass_type airmass;
	};
}

class storage:
	public table_storage< detail::sample > {
public:
	typedef storage self_type;
	typedef table_storage_column_getter<self_type, detail::sample::time_duration_type> time_column_type;
	typedef table_storage_column_getter<self_type, detail::sample::flux_type> flux_column_type;
	typedef table_storage_column_getter<self_type, detail::sample::airmass_type> airmass_column_type;

private:
	time_column_type m_time_column;
	flux_column_type m_flux_column;
	airmass_column_type m_airmass_column;

public:
	storage(size_type reserve = 60);

	const time_column_type& time() const noexcept { return m_time_column; }
	const flux_column_type& flux() const noexcept { return m_flux_column; }
	const airmass_column_type& airmass() const noexcept { return m_airmass_column; }

	void push_back(const time_duration_type& time, const flux_type& flux, airmass_type airmass) {
		table_storage< detail::sample >::push_back(detail::sample{time, flux, airmass});
	}
};

storage::storage(size_type reserve):
	table_storage<detail::sample>(reserve),
	m_time_column(*this, std::mem_fn(&detail::sample::time)),
	m_flux_column(*this, std::mem_fn(&detail::sample::flux)),
	m_airmass_column(*this, std::mem_fn(&detail::sample::airmass)) {}

} // flux

class worksheet {
	public:
		typedef utils::local_context<worksheet> context_type;
		typedef atmos::parameters::container    parameters_container_type;
		typedef atmos::parameters::proxy        parameters_proxy_type;
		typedef atmos::parameters::vector_type  vector_type;
		typedef atmos::background::factory      background_factory_type;

		constexpr static std::size_t window = 2;
	private:
		class input:
			public atmos::input::file,
			public utils::local_context<worksheet> {
		private:
			class measurement_accumulator:
				public utils::local_context< worksheet > {
			private:
				time_type m_starttime;
			public:
				measurement_accumulator( const time_type& time, context_reference_type ctx ) noexcept : local_context_type( ctx ), m_starttime(time) {};
				virtual ~measurement_accumulator() = 0;
				virtual void accumulate( const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) = 0;
				virtual void flush() = 0;

				template<class T> bool type() const {
					return dynamic_cast<const T*>(this);
				}

				inline const time_type& starttime() const { return m_starttime; }
			};

			class dummy_accumulator:
				public measurement_accumulator {
			public:
				dummy_accumulator( const time_type& time, context_reference_type ctx ) noexcept : measurement_accumulator( time, ctx ) {};
				virtual ~dummy_accumulator() = default;
				virtual void accumulate(  const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) {}
				virtual void flush() {}
			};

			class background_accumulator;
			class normal_accumulator;

		private:
			template<class T> void switch_accumulator( const time_type& time ) {
				std::unique_ptr< measurement_accumulator > new_accumulator( new T( time,context() ) );
				std::swap( m_maccum, new_accumulator );
				new_accumulator->flush();
			}

			template<class T> bool accumulator_type() const {
				return m_maccum->type<T>();
			}

		public:
			input( std::istream& stm, context_reference_type ctx ):
				atmos::input::file(stm),
				local_context_type(ctx),
				m_maccum(new dummy_accumulator(time_type(), ctx)) {}

			virtual ~input() = default;

			void on_comment( const std::string& line );
			void on_parameter( const time_type& time, const std::string& key, const std::string& value );
			void on_object( const time_type& time, const object_type& obj );
			void on_background( const time_type& time );
			void on_normal( const time_type& time );
			void on_flux( const time_type& time );
			void on_unknown_mode( const time_type& time );
			void on_measurements( const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 );
			void on_dimm( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars );

		private:
			std::unique_ptr<measurement_accumulator> m_maccum;
		};

		class output: public atmos::output::file {
		public:
			output(std::ostream& stm);

			void magnitude(const flux::time_type& time, const std::string& method, double mag, double mag_err, const vector_type& ch_mag);
			void magnitude(const flux::time_type& time, const std::string& method, double mag, double mag_err, const vector_type& ch_mag, const vector_type& ch_mag_err);
			void rconst(const flux::time_type& time, const std::string& method, const vector_type& rconst);
		};

		class object;

	private:
		atmos::background::model::abstract* select_background_model_smart( const atmos::background::time_type& time ) const;
		atmos::background::model::abstract* select_background_model_simple( const atmos::background::time_type& time ) const;

	private:
		const boost::program_options::variables_map& m_va;

		parameters_container_type m_pcontainer;
		parameters_proxy_type     m_pproxy;
		vector_type               m_rconst;
		background_factory_type   m_background_factory;

		input m_file_input;
		output m_file_output;

		std::unique_ptr<object> m_prevobject, m_curobject;

		void process_pair(const prephot::flux::time_type& time);

	public:
		static int parse_program_options(int argc, char** argv, boost::program_options::variables_map& va);

		worksheet(const boost::program_options::variables_map& va);
		~worksheet() = default;

		inline const parameters_proxy_type& parameters() const { return m_pproxy; }
		const vector_type& rconst() const noexcept { return m_rconst; }
		void advance_rconst(const vector_type& next_rconst) noexcept;
		void update_rconst(std::size_t channel, double rconst);
		double aperture_area() const;
		inline const background_factory_type& background_factory() const { return m_background_factory; }

		double sun_altitude(const boost::posix_time::ptime& time) const;

		inline object& curobject() {
			if (!m_curobject)
				throw std::logic_error("There is no any object");
			return *m_curobject;
		}

		void pre();
		void run();
		void post();
};

worksheet::output::output(std::ostream& stm):
	atmos::output::file(stm) {}

void worksheet::output::magnitude(const flux::time_type& time, const std::string& method, double mag, double mag_err, const vector_type& ch_mag) {
	assert(ch_mag.size() == atmos::channels);

	boost::io::ios_all_saver asaver( m_stm );

	m_stm << "X " << time << " " << method << std::fixed << " " << atmos::datum(mag, mag_err);
	for (auto it = ch_mag.begin(); it != ch_mag.end(); ++it) {
		m_stm << " " << atmos::datum(*it);
	}
	m_stm << std::endl;
}

void worksheet::output::magnitude(const flux::time_type& time, const std::string& method, double mag, double mag_err, const vector_type& ch_mag, const vector_type& ch_mag_err) {
	assert(ch_mag.size() == atmos::channels);
	assert(ch_mag_err.size() == atmos::channels);

	boost::io::ios_all_saver asaver( m_stm );

	m_stm << "X " << time << " " << method << std::fixed << " " << atmos::datum(mag, mag_err);
	for (std::size_t i = 0; i < BOOST_UBLAS_SAME(ch_mag.size(),ch_mag_err.size()); ++i) {
		m_stm << " " << atmos::datum(ch_mag(i), ch_mag_err(i));
	}
	m_stm << std::endl;
}

void worksheet::output::rconst(const flux::time_type& time, const std::string& method, const vector_type& rconst) {
	assert(rconst.size() == atmos::channels);

	boost::io::ios_all_saver asaver( m_stm );

	m_stm << "R " << time << " " << method << std::fixed;
	for (auto it = rconst.begin(); it != rconst.end(); ++it) {
		m_stm << " " << atmos::datum(*it);
	}
	m_stm << std::endl;
}

class worksheet::object:
	public utils::local_context< worksheet > {
public:
	typedef context_type::input::time_type time_type;

private:
	time_type m_seen;
	unsigned int m_hd;
	yalpa::equatorial<double> m_coords;
	double m_mag;

	flux::storage m_fluxes;
public:
	object(time_type seen, unsigned int hd, const yalpa::equatorial<double> ecoords, double mag, context_reference_type ctx):
		local_context_type(ctx),
		m_seen(seen),
		m_hd(hd),
		m_coords(ecoords),
		m_mag(mag) {}

	time_type    seen() const noexcept { return m_seen; }
	unsigned int hd()   const noexcept { return m_hd; }
	double       mag()  const noexcept { return m_mag; }

	flux::storage& storage() noexcept { return m_fluxes; }

	double airmass( const time_type& time ) const;
	bool operator==(const object& other) const noexcept;
	bool operator!=(const object& other) const noexcept;
};

worksheet::input::measurement_accumulator::~measurement_accumulator() = default;

class worksheet::input::background_accumulator:
	public worksheet::input::measurement_accumulator {
private:
	std::unique_ptr< atmos::background::model::abstract > m_model;
	const atmos::parameters::vector_type m_deadtime;
	const atmos::parameters::vector_type m_rconst;
	const double m_aperture_factor;
	const double m_exposition;	
	atmos::background::filter::nonpoisson m_nonpoisson_filter;
public:
	background_accumulator( const time_type& time, context_reference_type ctx ):
		measurement_accumulator(time, ctx),
		m_model(ctx.m_background_factory.create_new_model(time)),
		m_deadtime(ctx.parameters().nonlinearity()),
		m_rconst(ctx.rconst()),
		m_aperture_factor(mag_ln_const * std::log(ctx.aperture_area())),
		m_exposition(ctx.parameters().exposition()),
		m_nonpoisson_filter(
			ctx.m_va["background-threshold"].as<double>(),
			ctx.parameters().nonpoisson()(3),
			1000 * ctx.parameters().basetime() / ctx.parameters().exposition()) {

		if (!m_model)
			throw std::logic_error("There is no any background model");
	};
	virtual ~background_accumulator() {};
	virtual void accumulate(const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2) {
		using atmos::indices::flux_nonlinearity_correction;
		
		if (!m_nonpoisson_filter(time, mean, vars)) {
			if (m_model) {
				m_model->accumulate(time, flux_nonlinearity_correction(mean / m_exposition, m_deadtime));
			}
		}
	}
	virtual void flush() {
		using namespace boost::numeric::ublas;
		using namespace utils::algo;

		context().m_file_output.background(starttime());

		if (m_model && m_model->size()) {
			context().m_background_factory.handle_new_model(m_model.release());

			const vector_type background_flux = context().m_background_factory.background(starttime());
			const vector_type background_mag = -mag_ln_const * utils::ublas::log(background_flux) + m_rconst + scalar_vector<double>(atmos::channels, m_aperture_factor);
			const double mean_background_mag = mean(background_mag.begin(), background_mag.end());
			const double mean_background_mag_err = error_of_mean(background_mag.begin(), background_mag.end(), mean_background_mag);

			context().m_file_output.fluxes(starttime(), background_flux);
			context().m_file_output.magnitude(starttime(), "B", mean_background_mag, mean_background_mag_err, background_mag);
		}
	}
};

class worksheet::input::normal_accumulator:
	public worksheet::input::measurement_accumulator {
public:
	typedef context_type::input::size_type size_type;
private:
	time_type m_accumulation_begin;
	time_type m_accumulation_end;

	flux::storage m_fluxes;

	const atmos::parameters::vector_type m_deadtime;
	const atmos::parameters::vector_type m_scatter;
	const atmos::parameters::vector_type m_npoisson;
	const atmos::parameters::vector_type m_rconst;
	const double m_exposition;	

	atmos::indices::filter::flux_threshold m_flux_filter;
public:
	normal_accumulator( const time_type& time, context_reference_type ctx ) noexcept:
		measurement_accumulator(time, ctx),
		m_deadtime(ctx.parameters().nonlinearity()),
		m_scatter(ctx.parameters().scatter()),
		m_npoisson(ctx.parameters().nonpoisson()),
		m_rconst(ctx.rconst()),
		m_exposition(ctx.parameters().exposition()),
		m_flux_filter(ctx.m_va["flux-threshold"].as<double>()) {};

	virtual ~normal_accumulator() = default;
	virtual void accumulate(const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2) {
		/* mean is in photons
		   flux is in photons/msec */
		using namespace atmos::indices;

		if( m_accumulation_begin == boost::posix_time::not_a_date_time )
			m_accumulation_begin = time;
		m_accumulation_end = time;

		const flux_type flux = mean / m_exposition;
		flux_type flux_corrected = flux_nonlinearity_correction(flux, m_deadtime);

		const std::tuple< vars_type, rho1_type, rho2_type > vars_corrected = vars_nonlinearity_correction(
			flux_corrected,
			element_prod(vars, outer_prod(flux, flux)),
			element_prod(rho1, outer_prod(flux, flux)),
			element_prod(rho2, element_prod(flux, flux)), m_deadtime, m_npoisson, m_exposition);

		flux_corrected = flux_background_correction(flux_corrected, m_scatter, context().background_factory().background(time));

		if (!m_flux_filter(flux_corrected)) {
			const double airmass = context().curobject().airmass(time);

			m_fluxes.push_back(time - m_accumulation_begin, flux_corrected, airmass);
		}
	}
	virtual void flush() {
		using namespace boost::numeric::ublas;
		using namespace utils::algo;
		using namespace utils::ublas;

		const flux::airmass_type airmass_avg = mean(m_fluxes.airmass().begin(), m_fluxes.airmass().end());

		context().m_file_output.normal(starttime(), airmass_avg);

		if (m_fluxes.size() > 1) {
			const boost::posix_time::ptime meantime = m_accumulation_begin + (m_accumulation_end-m_accumulation_begin)/2;

			const flux::flux_type flux_avg = mean(m_fluxes.flux().begin(), m_fluxes.flux().end());
			const flux::flux_type flux_err = error_of_mean(m_fluxes.flux().begin(), m_fluxes.flux().end(), flux_avg);
			const vector_type extinction = (-mag_ln_const * utils::ublas::log(flux_avg) - scalar_vector<double>(atmos::channels, context().curobject().mag()) + m_rconst) / airmass_avg;
			const vector_type extinction_err = abs(element_div(flux_err, flux_avg) * (mag_ln_const / airmass_avg));
			const vector_type weights = inv(flux_err);
			const double mean_extinction = weighted_mean(extinction.begin(), extinction.end(), weights.begin(), weights.end());
			const double mean_extinction_err = error_of_weighted_mean(extinction.begin(), extinction.end(), weights.begin(), weights.end(), mean_extinction);

			context().m_file_output.fluxes(meantime, flux_avg, flux_err);
			context().m_file_output.magnitude(meantime, "S", mean_extinction, mean_extinction_err, extinction, extinction_err);

			context().curobject().storage().push_back(meantime - context().curobject().seen(), flux_avg, airmass_avg);

			if (context().curobject().storage().size() >= window) {
				context().process_pair(m_accumulation_end);
			}
		}
	}
};

double worksheet::object::airmass( const time_type& time ) const {
	yalpa::horizon<double> hc( yalpa::make_horizon_helper( m_coords, context().m_pproxy.site(), yalpa::sidereal_time(time) ) );
	double secz1 = 1.0/( std::cos(hc.coaltitude()) );
	double secz2 = secz1 - 1.0;
	return secz1 - secz2*( 0.0018167 + secz2*( 0.002875 + 0.0008083*secz2 ) );
}

bool worksheet::object::operator==(const object& other) const noexcept {
	return m_hd == other.m_hd;
}

bool worksheet::object::operator!=(const object& other) const noexcept {
	return !(*this == other);
}

void worksheet::input::on_comment( const std::string& line ) {
	/* do nothing */
}

void worksheet::input::on_parameter( const time_type& time, const std::string& key, const std::string& value ) {
	constexpr const char* prefix = "channels/rconstant/channel";

	context().m_pcontainer[key] = value;

	const auto it = std::mismatch(key.cbegin(), key.cend(), prefix, [] (const char x, const char y) noexcept {
		return std::tolower(x) == y;
	});

	if (*it.second == '\0' && it.first != key.cend()) {
		const std::size_t channel = std::tolower(*it.first) - 'a';
		const auto rconst = boost::lexical_cast<double>(value);

		context().update_rconst(channel, rconst);
	}

	context().m_file_output.parameter(time, key, value);
}

void worksheet::input::on_object( const time_type& time, const object_type& obj ) {
	if (!accumulator_type<dummy_accumulator>()) {
		switch_accumulator<dummy_accumulator>(time);
	}

	std::unique_ptr<context_type::object> nextobject(new context_type::object(time, obj.hd, obj.coordinates, obj.vmag, context()));

	if (!context().m_curobject) {
		context().m_curobject = std::move(nextobject);
	} else if (*context().m_curobject != *nextobject) {
		context().process_pair(time);

		std::swap(context().m_prevobject, context().m_curobject);
		std::swap(context().m_curobject, nextobject);
	}

	context().m_file_output.objectinfo(time, obj);
}

void worksheet::input::on_background( const time_type& time ) {
	switch_accumulator<background_accumulator>(time);
}

void worksheet::input::on_normal( const time_type& time ) {
	switch_accumulator<normal_accumulator>(time);
}

void worksheet::input::on_flux( const time_type& time ) {
	switch_accumulator<dummy_accumulator>(time);
}

void worksheet::input::on_unknown_mode( const time_type& time ) {
	switch_accumulator<dummy_accumulator>(time);
}

void worksheet::input::on_measurements( const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) {
	m_maccum->accumulate(time, mean, vars, rho1, rho2);
}

void worksheet::input::on_dimm( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars ) {
	/* ignore DIMM */
}

atmos::background::model::abstract* worksheet::select_background_model_smart( const atmos::background::time_type& time ) const {
	using namespace std::placeholders;
	using namespace atmos::background::model;
	
	double alt = sun_altitude( time ) * 180 / M_PI; // degrees

	if( alt > -5.0 )
		return new daytime( time );
	else if( alt > -12.0 )
		return new twilight( time, std::bind(std::mem_fn(&worksheet::sun_altitude), this, _1) );
	// else
	return new night( time );
}

atmos::background::model::abstract* worksheet::select_background_model_simple( const atmos::background::time_type& time ) const {
	using namespace atmos::background::model;
	return new night( time );
}

void worksheet::process_pair(const prephot::flux::time_type& time) {
	using namespace boost::numeric::ublas;
	using namespace utils::algo;

	constexpr flux::airmass_type airmass_limit = 0.25;
	const boost::posix_time::time_duration time_limit(0, 6, 0, 0); // 6 min

	if (!m_curobject || !m_prevobject)
		return;

	const flux::storage& cur_fluxes = m_curobject->storage();
	const flux::storage& prev_fluxes = m_prevobject->storage();

	const std::size_t cur_window  = std::min(window, cur_fluxes.size());
	const std::size_t prev_window = std::min(window, prev_fluxes.size());

	if (!cur_window || !prev_window)
		return;

	const flux::time_type    cur_time    = m_curobject->seen() + mean(cur_fluxes.time().begin(), cur_fluxes.time().begin() + cur_window);
	const flux::flux_type    cur_flux    = mean(cur_fluxes.flux().begin(),    cur_fluxes.flux().begin() + cur_window);
	const flux::airmass_type cur_airmass = mean(cur_fluxes.airmass().begin(), cur_fluxes.airmass().begin() + cur_window);

	const flux::time_type    prev_time    = m_prevobject->seen() + mean(prev_fluxes.time().end() - prev_window, prev_fluxes.time().end());
	const flux::flux_type    prev_flux    = mean(prev_fluxes.flux().end() - prev_window,    prev_fluxes.flux().end());
	const flux::airmass_type prev_airmass = mean(prev_fluxes.airmass().end() - prev_window, prev_fluxes.airmass().end());

	const boost::posix_time::time_duration diff_time = cur_time - prev_time;
	const flux::airmass_type diff_airmass = cur_airmass - prev_airmass;

	if (std::abs(diff_airmass) > airmass_limit && diff_time < time_limit) {
		const flux::flux_type cur_mag  = -mag_ln_const * utils::ublas::log(cur_flux)  - scalar_vector<double>(atmos::channels, m_curobject->mag());
		const flux::flux_type prev_mag = -mag_ln_const * utils::ublas::log(prev_flux) - scalar_vector<double>(atmos::channels, m_prevobject->mag());

		const vector_type extinction = (cur_mag - prev_mag) / diff_airmass;
		const vector_type rconst = (prev_airmass * cur_mag - cur_airmass * prev_mag) / diff_airmass;
		const vector_type expected_extinction = (-mag_ln_const * utils::ublas::log(cur_flux) - scalar_vector<double>(atmos::channels, m_curobject->mag()) + this->rconst()) / cur_airmass;

		const double mean_extinction = mean(extinction.begin(), extinction.end());
		const double mean_extinction_err = error_of_mean(extinction.begin(), extinction.end(), mean_extinction);
		const double mean_expected_extinction = mean(expected_extinction.begin(), expected_extinction.end());

		m_file_output.magnitude(cur_time, "P", mean_extinction, mean_extinction_err, extinction);

		m_file_output.rconst(cur_time, "P", rconst);

		if (mean_extinction > 0.0 && std::abs(mean_extinction - mean_expected_extinction) < 0.1) {
			advance_rconst(rconst);
			m_file_output.rconst(time, "S", m_rconst);
		}
	}

	m_prevobject.reset();
}

int worksheet::parse_program_options( int argc, char** argv, boost::program_options::variables_map& va ) {
	namespace po = boost::program_options;

	po::options_description desc("Allowed options");
	desc.add_options()
		( "background-threshold,b", po::value<double>()->default_value(4.0), "set background threshold" )
		( "enable-simple-background", "do not account for the sun's altitude")
		( "flux-threshold",  po::value<double>()->default_value(2.0), "set flux threshold" )
		( "help,h",    "produce help message" )
		( "version,v", "produce version message" );

	try {
		po::store( po::parse_command_line( argc, argv, desc ), va );
		po::notify( va );
		if( va.count("version") ) {
			std::cerr << (boost::format("prephot %1%") % VERSION) << std::endl;
			std::cerr << "Copyright (C) 2009-2025 Sternberg Astronomical Institute, MSU" << std::endl;
			std::cerr << "This is free software; see the source for copying conditions.  There is NO" << std::endl;
			std::cerr << "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE." << std::endl;
			return 1;
		}
		if( va.count("help") ) {
			std::cerr << desc << std::endl;
			return 1;
		}
	} catch ( const po::error& e ) {
		std::cerr << e.what() << std::endl;
		std::cerr << desc << std::endl;
		return 1;
	}
	return 0;
}

worksheet::worksheet( const boost::program_options::variables_map& va ):
	m_va(va),
	m_pcontainer(),
	m_pproxy(m_pcontainer),
	m_rconst(atmos::channels),
	m_background_factory(std::bind(std::mem_fn(va.count("enable-simple-background") ? &worksheet::select_background_model_simple : &worksheet::select_background_model_smart ), this, std::placeholders::_1)),
	m_file_input(std::cin, *this),
	m_file_output(std::cout) {

	/* Default instrumental constant */
	m_rconst(0) = 6.0;
	m_rconst(1) = 6.7;
	m_rconst(2) = 8.0;
	m_rconst(3) = 8.4;
}

void worksheet::advance_rconst(const vector_type& next_rconst) noexcept {
	m_rconst = 0.93 * m_rconst + 0.07 * next_rconst;
}

void worksheet::update_rconst(std::size_t channel, double rconst) {
	if (channel >= m_rconst.size()) {
		throw std::out_of_range("Wrong channel index");
	}

	m_rconst[channel] = rconst;
}

double worksheet::aperture_area() const {
	constexpr const char* field_aperture_key = "Camera/Geometry/FieldAperture";
	constexpr const char* scale_key = "Camera/Geometry/Scale";

	if (! (m_pcontainer.has(field_aperture_key) && m_pcontainer.has(scale_key))) {
		return 39700.0; // some meaningful default: 3.75 arcmin FOV
	}

	const auto fov = m_pcontainer.as<double>(field_aperture_key) * m_pcontainer.as<double>(scale_key);

	return M_PI * (fov * fov);
}

double worksheet::sun_altitude( const boost::posix_time::ptime& time ) const {
	yalpa::ecliptic<double> sun = yalpa::sun_position(time);
	yalpa::horizon<double> sun_hz( make_horizon_helper(sun, m_pproxy.site(), yalpa::sidereal_time(time)));
	return 0.5 * M_PI - sun_hz.coaltitude();
}

void worksheet::pre() {
	/* do nothing */
	m_file_output.comment((boost::format("prephot %1%") % VERSION).str());
}
void worksheet::run() {
	bool ret = true;
	while( ret ) {
		try {
			ret &= m_file_input.shift();
		} catch ( const std::exception& e ) {
			std::cerr << m_file_input.line() << ": " << e.what() << std::endl;
		}
	}
}
void worksheet::post() {
}

} // prephot

int main(int argc, char **argv) {
	using prephot::worksheet;

	boost::program_options::variables_map va;

	if (worksheet::parse_program_options(argc, argv, va)) {
		return 1;
	}

	try {
		worksheet w(va);
		
		w.pre();
		w.run();
		w.post();

	} catch ( const std::exception& e ) {
		std::cerr << e.what() << std::endl;
		return 1;
	}

	return 0;
}
